# nvcomp 2.2.0 (2022-02-07)
## New Features
 - Entropy-only mode for GDeflate
 - New high-level interface
 - Windows support
 - Support for GPU-accelerated ANS

## Interface Changes

### High level interface
 - High level interface is now standardized across compressor formats. 
 - This interface provides a single nvcompManagerBase object that
 can do compression and decompression. Users can now decompress nvcomp-compressed
 files without knowing how they were compressed. The interface also can manage scratch space 
 and splitting the input buffer into independent chunks for parallel processing.

### API Consolidation
 - nvCOMP now supports only the low-level batch API and the new high level interface

# nvcomp 2.1.0 (2021-10-28)

## New Features
 - New release of low-level batched API for Cascaded and Bitcomp methods.
 - New high-throughput and high-compression-ratio GPU compressors in GDeflate

## Interface Changes
 - Update batched/low-level compression interfaces to take an options parameter,
 to allow configuring future compression algorithms.
 - Update batched/low-level decompression interfaces to output the decompressed
 size (or 0 if an error occurs).
 - Add bounds checking to batched/low-level decompression routines, such that
 if an invalid compressed data stream is provided, 0 will be written for the
 output size, rather than generating an illegal memory access.
 - Fix LZ4 to support chunk sizes < 32 KB. 

## Performance Optimizations
 - Improve performance of Snappy compression by ~10% in some configurations.
 - Add an optimization to the LZ4 compressor based on specification of input data as
 char, short, or int, rather than just treating the input as raw bytes. 
 - Optimization to reduce the LZ hash table size when compressing smaller chunks.
 - Improved compression performance in GDeflate with the high-throughput option
 - Improved decompression performance in GDeflate (10-75% depending on the dataset)

## Bug Fixes
 - Fix LZ4 CPU compression example.
 - Fix temp allocation size bug in `benchmark_template_chunked`.

## Infrastructure
 - Update CMakeLists to compile nvcomp with -fPIC enabled.
 - Add a new script for benchmarking compression algorithms.
 - Add unit tests for the Snappy decompressor that tests decompression on legally
 formatted files that won't be generated by the nvcomp compressor due to 
 configuration.
 - Update CMakeLists to suppress warnings about missing nvcomp external dependencies
 when the user didn't indicate they wanted to include them.
 - Update CMakeLists to allow install into include folder that the user does not have 
 ownership of.
 
# nvcomp 2.0.2 (2021-06-30)

- Add example `lz4_cpu_decompression` to compress on the GPU with nvCOMP and
  decompress on the CPU with `liblz4`.
- Add CMake option for building a static library.
- Fix bug in LZ4 compression kernel to comply with LZ4 end of block
  restrictions.
- Fix temp allocation size bug in `benchmark_lz4_chunked`.

# nvcomp 2.0.1 (2021-06-08)

- Improve CMake setup for using nvCOMP as a submodule. This includes marking
dependencies as PRIVATE, and adding options for building examples, tests, and
benchmarks (e.g., `-DBUILD_EXAMPLES=ON`, `-DBUILD_TESTS=ON`, and
`-DBUILD_BENCHMARKS=ON`).
- Fix double free error in `benchmark_snappy_synth`.
- Fix copy direction in Cascaded compression when the output size on the GPU.
- Improve testing coverage.
- Mark the generic decompression interfaces defined in `include/nvcomp.h` as
deprecated.

# nvcomp 2.0.0 (2021-04-28)

- Replace previous C, and C++ APIs.
- Added Snappy compression (batched interface).
- Added support for using Bitcomp and GDeflate external compressors.
- Added `/examples` folder demonstrating use cases interface with CPU
implementations of LZ4 and GDeflate, as well as GPU Direct Storage.
- Improve support for Windows in benchmark implementations.
- Made usage of `std::uniform_int_distribution<>` in the benchmarks
conform to the C++14 standard.
- Fix issue in Cascaded compression when using the default configuration
('auto'), for small inputs.

# nvcomp 1.2.3 (2021-04-07)

- Fix bug in LZ4 compression kernel for the Pascal architecture.

# nvcomp 1.2.2 (2021-02-08)

- Fix linking errors in Clang++.
- Fix error being incorrectly returned by Cascaded compression when output memory was initialized to
all `-1`'s.
- Fix C++17 style static assert.
- Fix prematurely freeing memory in Cascaded compression.
- Fix input format and usage messaging for benchmarks.

# nvcomp 1.2.1 (2020-12-21)

- Fix compile error and unit tests for cascaded selector.

# nvcomp 1.2.0 (2020-12-19)

- Add the Cascaded Selector and Cascaded Auto set of interfaces for
automatically configuring cascaded compression.
- Generally improve error handling and messaging.
- Update CMake configuration to support CCache.

# nvcomp 1.1.1 (2020-12-02)

- Add all-gather benchmark.
- Add sm80 target if CUDA version is 11 or greater.

# nvcomp 1.1.0 (2020-10-05)

- Add batch C interface for LZ4, allowing compressing/decompressing multiple
inputs at once.
- Significantly improve performance of LZ4 compression.

# nvcomp 1.0.2 (2020-08-12)

- Fix metadata freeing for LZ4, to avoid possible mismatch of `new[]` and
`delete`.


# nvcomp 1.0.1 (2020-08-07)

- Fixed naming of nvcompLZ4CompressX functions in `include/lz4.h`, to have the
`nvcomp` prefix.
- Changed CascadedMetadata::Header struct initialization to work around
internal compiler error.


# nvcomp 1.0.0 (2020-07-31)

- Initial public release.

